from pathlib import Path
import sys
import unittest
from unittest.mock import patch

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))

from pyodoo_rpc_client import OdooRpcClient
from pyodoo_rpc_client.entity import OdooRpcEntity
from pyodoo_rpc_client.exceptions import OdooRpcError


class DummyCommonProxy:
    def __init__(self, uid=1):
        self.uid = uid
        self.calls = []

    def authenticate(self, db, username, password, context):
        self.calls.append({"db": db, "username": username, "password": password, "context": context})
        return self.uid


class DummyModelsProxy:
    def __init__(self, responses):
        self.responses = list(responses)
        self.calls = []

    def execute_kw(self, db, uid, password, model_name, method, args, kwargs):
        self.calls.append(
            {
                "db": db,
                "uid": uid,
                "password": password,
                "model_name": model_name,
                "method": method,
                "args": args,
                "kwargs": kwargs,
            }
        )
        item = self.responses.pop(0)
        if isinstance(item, Exception):
            raise item
        return item


class OdooRpcClientBehaviorTests(unittest.TestCase):
    def test_model_search_read_merges_context(self):
        common_proxy = DummyCommonProxy(uid=42)
        models_proxy = DummyModelsProxy(responses=[[{"id": 7, "name": "Acme"}]])

        with patch("xmlrpc.client.ServerProxy", side_effect=[common_proxy, models_proxy]):
            client = OdooRpcClient(
                url="https://example.com",
                db="testdb",
                username="api@example.com",
                password="secret",
            )

        model = client.model("res.partner").with_context({"lang": "en_US"})
        result = model.search_read([["is_company", "=", True]], context={"tz": "UTC"})

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], OdooRpcEntity)
        self.assertEqual(result[0].id, 7)

        call = models_proxy.calls[0]
        self.assertEqual(call["method"], "search_read")
        self.assertEqual(call["args"][0], ["is_company", "=", True])
        self.assertEqual(call["kwargs"]["context"]["lang"], "en_US")
        self.assertEqual(call["kwargs"]["context"]["tz"], "UTC")

    def test_execute_kw_returns_default_when_debug_false_on_rpc_error(self):
        common_proxy = DummyCommonProxy(uid=1)
        models_proxy = DummyModelsProxy(responses=[RuntimeError("boom")])

        with patch("xmlrpc.client.ServerProxy", side_effect=[common_proxy, models_proxy]):
            client = OdooRpcClient(
                url="https://example.com",
                db="testdb",
                username="api@example.com",
                password="secret",
                debug=False,
            )

        default_value = []
        result = client.execute_kw("res.partner", "search_read", args=[[]], kwargs={}, default=default_value)

        self.assertEqual(result, default_value)
        self.assertIsInstance(client.error, OdooRpcError)



from __future__ import annotations

import copy
from typing import Any, Dict, Optional
from typing import TYPE_CHECKING

from .entity import OdooRpcEntity

if TYPE_CHECKING:
    from .client import OdooRpcClient


class OdooRpcModel:
    def __init__(
        self,
        client: "OdooRpcClient",
        model_name: str,
        context: Optional[Dict[str, Any]] = None,
        debug: Optional[bool] = None,
    ):
        self.client = client
        self.model = model_name
        self.context = copy.deepcopy(context or {})
        self.error = None
        self._fields = None
        self._debug = client.debug if debug is None else bool(debug)

    def set_debug(self, debug: bool):
        self._debug = bool(debug)
        return self

    def with_context(self, context: Optional[Dict[str, Any]] = None):
        merged = copy.deepcopy(self.context)
        if context:
            merged.update(copy.deepcopy(context))
        return self.__class__(
            client=self.client,
            model_name=self.model,
            context=merged,
            debug=self._debug,
        )

    def fields(self):
        if self._fields is None:
            data = self.fields_get([], {"attributes": ["string", "help", "type", "required", "relation"]})
            self._fields = data if isinstance(data, dict) else {}
        return self._fields

    @staticmethod
    def _normalize_ids(ids: Any):
        if ids is None:
            return []
        if isinstance(ids, (int, str)):
            return [int(ids)]
        if isinstance(ids, tuple):
            ids = list(ids)
        if isinstance(ids, list):
            if len(ids) == 1 and isinstance(ids[0], list):
                ids = ids[0]
            parsed = []
            for item in ids:
                try:
                    parsed.append(int(item))
                except Exception:
                    continue
            return parsed
        return ids

    @staticmethod
    def _normalize_domain(domain: Any):
        if isinstance(domain, list) and len(domain) == 1 and isinstance(domain[0], list):
            return domain[0]
        return domain

    @staticmethod
    def _default_method_return(method: str):
        method = method.lower()
        if method in {"search", "search_read", "read"}:
            return []
        if method in {"fields_get"}:
            return {}
        if method in {"write", "unlink"}:
            return False
        if method in {"create"}:
            return None
        return []

    def _build_rpc_call(self, method: str, args: tuple, kwargs: dict):
        call_args = list(args)
        call_kwargs = copy.deepcopy(kwargs)
        lower_method = method.lower()

        if lower_method in {"search", "search_read"}:
            if call_args:
                call_args[0] = self._normalize_domain(call_args[0])
        elif lower_method == "read":
            if call_args:
                call_args[0] = self._normalize_ids(call_args[0])
        elif lower_method == "create":
            if call_args:
                first = call_args[0]
                if isinstance(first, tuple):
                    first = list(first)
                if isinstance(first, list) and len(first) == 1 and isinstance(first[0], dict):
                    first = first[0]
                call_args[0] = first
        elif lower_method == "write":
            ids_arg = call_args[0] if len(call_args) >= 1 else None
            vals_arg = call_args[1] if len(call_args) >= 2 else None

            if len(call_args) == 1 and isinstance(call_args[0], (list, tuple)) and len(call_args[0]) >= 2:
                ids_arg = call_args[0][0]
                vals_arg = call_args[0][1]

            call_args = []
            if ids_arg is not None:
                call_args.append(self._normalize_ids(ids_arg))
            if vals_arg is not None:
                vals = vals_arg
                if isinstance(vals, list) and len(vals) == 1 and isinstance(vals[0], dict):
                    vals = vals[0]
                call_args.append(vals)
        elif lower_method == "unlink":
            if call_args:
                call_args[0] = self._normalize_ids(call_args[0])

        return call_args, call_kwargs

    def _exec(self, method: str, args=None, kwargs=None, default=None):
        kwargs = copy.deepcopy(kwargs or {})
        if self.context:
            ctx = copy.deepcopy(self.context)
            if isinstance(kwargs.get("context"), dict):
                ctx.update(kwargs["context"])
            kwargs["context"] = ctx
        result = self.client.execute_kw(self.model, method, args=args or [], kwargs=kwargs, debug=self._debug, default=default)
        self.error = self.client.error
        return result

    def execute(self, method: str, *args, **kwargs):
        default = self._default_method_return(method)
        try:
            call_args, call_kwargs = self._build_rpc_call(method, args, kwargs)
        except Exception as exc:
            self.error = exc
            if self._debug:
                raise
            return default
        return self._exec(method, args=call_args, kwargs=call_kwargs, default=default)

    def execute_kw(self, method: str, *args, **kwargs):
        return self.execute(method, *args, **kwargs)

    def search(self, domain, **kwargs):
        return self.execute("search", domain, **kwargs)

    def read(self, ids, **kwargs):
        return self.execute("read", ids, **kwargs)

    def search_read(self, domain, **kwargs):
        records = self.execute("search_read", domain, **kwargs)
        if not isinstance(records, list):
            return []
        return [OdooRpcEntity(self, data) for data in records]

    def create(self, vals, **kwargs):
        return self.execute("create", vals, **kwargs)

    def write(self, ids, vals, **kwargs):
        return self.execute("write", ids, vals, **kwargs)

    def unlink(self, ids, **kwargs):
        return self.execute("unlink", ids, **kwargs)

    def get(self, pk):
        return OdooRpcEntity(self, pk)

    def entity(self, data):
        return OdooRpcEntity(self, data)

    def new(self):
        return OdooRpcEntity(self)

    def __getattr__(self, method):
        def function(*_args, **_kwargs):
            return self.execute(method, *_args, **_kwargs)

        return function



from dataclasses import dataclass
from typing import Optional


@dataclass
class OdooRpcError(Exception):
    message: str
    cause: Optional[Exception] = None

    def __str__(self) -> str:
        return self.message


class OdooRpcConfigError(ValueError):
    pass



from pathlib import Path
import sys
import unittest

ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))


class PublicExportsTests(unittest.TestCase):
    def test_public_exports_importable(self):
        from pyodoo_rpc_client import OdooRpcClient, OdooRpcConfigError, OdooRpcEntity, OdooRpcError, OdooRpcModel

        self.assertIsNotNone(OdooRpcClient)
        self.assertIsNotNone(OdooRpcModel)
        self.assertIsNotNone(OdooRpcEntity)
        self.assertIsNotNone(OdooRpcError)
        self.assertIsNotNone(OdooRpcConfigError)



from __future__ import annotations

from typing import Any, Dict, Iterable, Optional

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .model import OdooRpcModel


class OdooRpcEntity:
    def __init__(self, model: "OdooRpcModel", data: Any = None):
        self.__model = model
        self.__init_data: Dict[str, Any] = {}
        self.__fields: Dict[str, Dict[str, Any]] = {}
        self.__id = None

        fields = model.fields()
        if fields:
            fields = dict(fields)
            fields.pop("id", None)
            self.__fields = fields

        if data is None:
            return

        if isinstance(data, dict):
            self.__init_data = dict(data)
            if data.get("id") is not None:
                self.__id = data["id"]
            self.set_data(data)
        else:
            self.load(data)

    @property
    def id(self):
        return self.__id

    @property
    def _model(self):
        return self.__model

    def exists(self) -> bool:
        return bool(self.__id)

    def __iter__(self):
        if self.__id is not None:
            yield "id", self.__id
        for key in self.__fields.keys():
            if hasattr(self, key):
                yield key, getattr(self, key)

    def load(self, record_id: Any):
        result = self.__model.read([int(record_id)])
        data = dict(result[0]) if result else {}
        self.__init_data = data
        if data.get("id") is not None:
            self.__id = data["id"]
        self.set_data(data)

    def refresh(self):
        if self.__id is not None:
            self.load(self.__id)

    def set_data(self, data: Dict[str, Any]):
        if not data:
            return
        item_data = dict(data)
        item_data.pop("id", None)
        for key, value in item_data.items():
            setattr(self, key, value)

    def get_data(self, fields: Optional[Iterable[str]] = None):
        data = dict(self)
        if fields:
            return {key: data[key] for key in fields if key in data or key == "id"}
        return data

    def get_changed_data(self):
        current_data = dict(self)
        return {key: val for key, val in current_data.items() if val != self.__init_data.get(key, None)}

    def get_changed_fields(self):
        return self.get_changed_data().keys()

    @staticmethod
    def _extract_id(create_result: Any):
        if isinstance(create_result, list):
            return create_result[0] if create_result else None
        return create_result

    def save(self):
        if self.__id is not None:
            data = self.get_changed_data()
            if data:
                self.__model.write([self.__id], data)
        else:
            create_result = self.__model.create(dict(self))
            created_id = self._extract_id(create_result)
            if created_id is not None:
                self.__id = created_id
                self.refresh()
        return self.__id

    def delete(self):
        if self.__id is not None:
            return self.__model.unlink([self.__id])
        return False



from __future__ import annotations

import xmlrpc.client
from typing import Any, Dict, Optional

from .exceptions import OdooRpcConfigError, OdooRpcError


class OdooRpcClient:
    def __init__(
        self,
        url: str,
        db: str,
        username: str,
        password: Optional[str] = None,
        key: Optional[str] = None,
        debug: bool = False,
        allow_none: bool = True,
    ):
        if not url:
            raise OdooRpcConfigError("url is required")
        if not db:
            raise OdooRpcConfigError("db is required")
        if not username:
            raise OdooRpcConfigError("username is required")

        self.url = str(url).rstrip("/")
        self.db = db
        self.username = username
        self.password = key or password
        self.debug = bool(debug)
        self.allow_none = allow_none

        self.uid = None
        self.logged_in = False
        self.error = None

        self.common = None
        self.models = None

        self._connect()

    def _connect(self):
        if not self.password:
            self.error = OdooRpcConfigError("password/key is required")
            self.logged_in = False
            return
        try:
            self.common = xmlrpc.client.ServerProxy(f"{self.url}/xmlrpc/2/common", allow_none=self.allow_none)
            self.uid = self.common.authenticate(self.db, self.username, self.password, {})
            self.models = xmlrpc.client.ServerProxy(f"{self.url}/xmlrpc/2/object", allow_none=self.allow_none)
            self.logged_in = bool(self.uid)
            self.error = None
        except Exception as exc:
            self.error = OdooRpcError("Failed to initialize RPC client", cause=exc)
            self.logged_in = False
            if self.debug:
                raise self.error

    def set_debug(self, debug: bool):
        self.debug = bool(debug)
        return self

    def model(self, model_name: str):
        from .model import OdooRpcModel

        return OdooRpcModel(client=self, model_name=model_name)

    def execute_kw(
        self,
        model_name: str,
        method: str,
        args: Optional[list] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        debug: Optional[bool] = None,
        default: Any = None,
    ):
        if not self.logged_in or self.uid is None:
            exc = OdooRpcError("Client is not authenticated")
            self.error = exc
            if self.debug if debug is None else bool(debug):
                raise exc
            return default

        try:
            result = self.models.execute_kw(
                self.db,
                self.uid,
                self.password,
                model_name,
                method,
                args or [],
                kwargs or {},
            )
            self.error = None
            return result
        except Exception as exc:
            wrapped = OdooRpcError(f"RPC call failed for {model_name}.{method}", cause=exc)
            self.error = wrapped
            should_raise = self.debug if debug is None else bool(debug)
            if should_raise:
                raise wrapped
            return default



from .client import OdooRpcClient
from .entity import OdooRpcEntity
from .exceptions import OdooRpcConfigError, OdooRpcError
from .model import OdooRpcModel

__all__ = [
    "OdooRpcClient",
    "OdooRpcModel",
    "OdooRpcEntity",
    "OdooRpcError",
    "OdooRpcConfigError",
]
